// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Program wcfs provides filesystem server with file data backed by wendelin.core arrays.
//
// Intro
//
// Each wendelin.core array (ZBigArray) is actually a linear file (ZBigFile)
// and array metadata like dtype, shape and strides associated with it. This
// program exposes as files only ZBigFile data and leaves rest of
// array-specific handling to clients. Every ZBigFile is exposed as one separate
// file that represents whole ZBigFile's data.
//
// For a client, the primary way to access a bigfile should be to mmap
// head/bigfile/<bigfileX> which represents always latest bigfile data.
// Clients that want to get isolation guarantee should subscribe for
// invalidations and re-mmap invalidated regions to file with pinned bigfile revision for
// the duration of their transaction. See "Invalidation protocol" for details.
//
// In the usual situation when bigfiles are big, and there are O(1)/δt updates,
// there should be no need for any cache besides shared kernel cache of latest
// bigfile data.
//
//
// Filesystem organization
//
// Top-level structure of provided filesystem is as follows:
//
//	head/			; latest database view
//		...
//	@<rev1>/		; database view as of revision <revX>
//		...
//	@<rev2>/
//		...
//	...
//
// where head/ represents latest data as stored in upstream ZODB, and
// @<revX>/ represents data as of database revision <revX>.
//
// head/ has the following structure:
//
//	head/
//		at			; data inside head/ is as of this ZODB transaction
//		watch			; channel for bigfile invalidations
//		bigfile/		; bigfiles' data
//			<oid(bigfile1)>
//			<oid(bigfile2)>
//			...
//
// where /bigfile/<bigfileX> represents latest bigfile data as stored in
// upstream ZODB. As there can be some lag receiving updates from the database,
// /at describes precisely ZODB state for which bigfile data is currently
// exposed. Whenever bigfile data is changed in upstream ZODB, information
// about the changes is first propagated to /watch, and only after that
// /bigfile/<bigfileX> is updated. See "Invalidation protocol" for details.
//
// @<revX>/ has the following structure:
//
//	@<revX>/
//		at
//		bigfile/		; bigfiles' data as of revision <revX>
//			<oid(bigfile1)>
//			<oid(bigfile2)>
//			...
//
// where /bigfile/<bigfileX> represent bigfile data as of revision <revX>.
//
// Unless accessed {head,@<revX>}/bigfile/<bigfileX> are not automatically visible in
// wcfs filesystem. Similarly @<revX>/ should be explicitly created by client via mkdir.
//
//
// Invalidation protocol
//
// In order to support isolation, wcfs implements invalidation protocol that
// must be cooperatively followed by both wcfs and client.
//
// First, client mmaps latest bigfile, but does not access it
//
//	mmap(head/bigfile/<bigfileX>)
//
// Then client opens head/watch and tells wcfs through it for which ZODB state
// it wants to get bigfile's view.
//
//	C: 1 watch <bigfileX> @<at>
//
// The server then, after potentially sending initial pin messages (see below),
// reports either success or failure:
//
//	S: 1 ok
//	S: 1 error ...		; if <at> is too far away back from head/at
//
// The server sends "ok" reply only after head/at is ≥ requested <at>, and
// only after all initial pin messages are fully acknowledged by the client.
// The client can start to use mmapped data after it gets "ok".
// The server sends "error" reply if requested <at> is too far away back from
// head/at.
//
// Upon watch request, either initially, or after sending "ok", the server will be notifying the
// client about file blocks that client needs to pin in order to observe file's
// data as of <at> revision:
//
// The filesystem server itself receives information about changed data from
// ZODB server through regular ZODB invalidation channel (as it is ZODB client
// itself). Then, separately for each changed file block, before actually
// updating head/bigfile/<bigfileX> content, it notifies through head/watch to
// clients, that had requested it (separately to each client), about the
// changes:
//
//	S: 2 pin <bigfileX> #<blk> @<rev_max>
//
// and waits until all clients confirm that changed file block can be updated
// in global OS cache.
//
// The client in turn should now re-mmap requested to be pinned block to bigfile@<rev_max>
//
//	# mmapped at address corresponding to #blk
//	mmap(@<rev_max>/bigfile/<bigfileX>, #blk, MAP_FIXED)
//
// and must send ack back to the server when it is done:
//
//	C: 2 ack
//
// The server sends pin notifications only for file blocks, that are known to
// be potentially changed after client's <at>, and <rev_max> describes the
// upper bound for the block revision:
//
//	<at>	<  <rev_max>
//
// The server maintains short history tail of file changes to be able to
// support openings with <at> being slightly in the past compared to current
// head/at. The server might reject a watch request if <at> is too far away in
// the past from head/at. The client is advised to restart its transaction with
// more uptodate database view if it gets watch setup error.
//
// A later request from the client for the same <bigfileX> but with different
// <at>, overrides previous watch request for that file. A client can use "-"
// instead of "@<at>" to stop watching a file.
//
// A single client can send several watch requests through single head/watch
// open, as well as it can use several head/watch opens simultaneously.
// The server sends pin notifications for all files requested to be watched via
// every head/watch open.
//
// Note: a client could use a single watch to manage its several views for the same
// file but with different <at>. This could be achieved via watching with
// @<at_min>, and then deciding internally which views needs to be adjusted and
// which views need not. Wcfs does not oblige clients to do so though, and a
// client is free to use as many head/watch openenings as it needs to.
//
// When clients are done with @<revX>/bigfile/<bigfileX> (i.e. client's
// transaction ends and array is unmapped), the server sees number of opened
// files to @<revX>/bigfile/<bigfileX> drops to zero, and automatically
// destroys @<revX>/bigfile/<bigfileX> after reasonable timeout.
//
//
// Protection against slow or faulty clients
//
// If a client, on purpose or due to a bug or being stopped, is slow to respond
// with ack to file invalidation notification, it creates a problem because the
// server will become blocked waiting for pin acknowledgments, and thus all
// other clients, that try to work with the same file, will get stuck.
//
// The problem could be avoided, if wcfs would reside inside OS kernel and this
// way could be able to manipulate clients address space directly (then
// invalidation protocol won't be needed). It is also possible to imagine
// mechanism, where wcfs would synchronously change clients' address space via
// injecting trusted code and running it on client side via ptrace to adjust
// file mappings.
//
// However ptrace does not work when client thread is blocked under pagefault,
// and that is exactly what wcfs would need to do to process invalidations
// lazily, because eager invalidation processing results in prohibitively slow
// file opens. See internal wcfs overview for details about why ptrace
// cannot be used and why lazy invalidation processing is required.
//
// Lacking OS primitives to change address space of another process and not
// being able to work it around with ptrace in userspace, wcfs takes approach
// to kill a slow client on 30 seconds timeout by default.
//
//
// Writes
//
// As each bigfile is represented by 1 synthetic file, there can be several
// write schemes:
//
// 1. mmap(MAP_PRIVATE) + writeout by client
//
// In this scheme bigfile data is mmapped in MAP_PRIVATE mode, so that local
// user changes are not automatically propagated back to the file. When there
// is a need to commit, client investigates via some OS mechanism, e.g.
// /proc/self/pagemap or something similar, which pages of this mapping it
// modified. Knowing this it knows which data it dirtied and so can write this
// data back to ZODB itself, without filesystem server providing write support.
//
// 2. mmap(MAP_SHARED, PROT_READ) + write-tracking & writeout by client
//
// In this scheme bigfile data is mmaped in MAP_SHARED mode with read-only pages
// protection. Then whenever write fault occurs, client allocates RAM from
// shmfs, copies faulted page to it, and then mmaps RAM page with RW protection
// in place of original bigfile page. Writeout implementation should be similar
// to "1", only here client already knows the pages it dirtied, and this way
// there is no need to consult /proc/self/pagemap.
//
// The advantage of this scheme over mmap(MAP_PRIVATE) is that in case
// there are several in-process mappings of the same bigfile with overlapping
// in-file ranges, changes in one mapping will be visible in another mapping.
// Contrary: whenever a MAP_PRIVATE mapping is modified, the kernel COWs
// faulted page into a page completely private to this mapping, so that other
// MAP_PRIVATE mappings of this file, including ones created from the same
// process, do not see changes made to the first mapping.
//
// Since wendelin.core needs to provide coherency in between different slices
// of the same array, this is the mode wendelin.core actually uses.
//
// 3. write to wcfs
//
// TODO we later could implement "write-directly" mode where clients would write
// data directly into the file.
package main

// Wcfs organization
//
// Wcfs is a ZODB client that translates ZODB objects into OS files as would
// non-wcfs wendelin.core do for a ZBigFile. Contrary to non-wcfs wendelin.core,
// it keeps bigfile data in shared OS cache efficiently. It is organized as follows:
//
// 1) 1 ZODB connection for "latest data" for whole filesystem (zhead).
// 2) head/bigfile/* of all bigfiles represent state as of zhead.At .
// 3) for head/bigfile/* the following invariant is maintained:
//
//	#blk ∈ file cache    =>    ZBlk(#blk) + all BTree/Bucket that lead to it  ∈ zhead cache
//	                           (ZBlk* in ghost state(%))
//
//    The invariant helps on invalidation: if we see a changed oid, and
//    zhead.cache.lookup(oid) = ø -> we know we don't have to invalidate OS
//    cache for any part of any file (even if oid relates to a file block - that
//    block is not cached and will trigger ZODB load on file read).
//
//    Currently we maintain this invariant by simply never evicting LOBTree/LOBucket
//    objects from ZODB Connection cache (LOBucket keeps references to ZBlk* and
//    so ZBlk* also stay in cache in ghost form). In the future we may want to
//    try to synchronize to kernel freeing its pagecache pages.
//
// 4) when we receive an invalidation message from ZODB - we process it and
//    propagate invalidations to OS file cache of head/bigfile/*:
//
//	invalidation message: (tid↑, []oid)
//
//    4.1) zhead.cache.lookup(oid)
//    4.2) ø: nothing to do - see invariant ^^^.
//    4.3) obj found:
//
//	- ZBlk*		-> file/#blk
//	- BTree/Bucket	-> δ(BTree)  -> file/[]#blk
//
//	in the end after processing all []oid from invalidation message we have
//
//	  [] of file/[]#blk
//
//	that describes which file(s) parts needs to be invalidated.
//
//    4.4) for all file/blk to invalidate we do:
//
//	- try to retrieve head/bigfile/file[blk] from OS file cache(*);
//	- if retrieved successfully -> store retrieved data back into OS file
//	  cache for @<rev>/bigfile/file[blk], where
//
//	    rev = max(δFtail.by(#blk)) || min(rev ∈ δFtail) || zhead.at	; see below about δFtail
//
//	- invalidate head/bigfile/file[blk] in OS file cache.
//
//	This preserves previous data in OS file cache in case it will be needed
//	by not-yet-uptodate clients, and makes sure file read of head/bigfile/file[blk]
//	won't be served from OS file cache and instead will trigger a FUSE read
//	request to wcfs.
//
//    4.5) no invalidation messages are sent to wcfs clients at this point(+).
//
//    4.6) processing ZODB invalidations and serving file reads (see 7) are
//      organized to be mutually exclusive.
//
// 5) after OS file cache was invalidated, we resync zhead to new database
//    view corresponding to tid.
//
// 6) for every file δFtail invalidation info about head/data is maintained:
//
//	- tailv: [](rev↑, []#blk)
//	- by:    {} #blk -> []rev↑ in tail
//
//    δFtail.tail describes invalidations to file we learned from ZODB invalidation.
//    δFtail.by   allows to quickly lookup information by #blk.
//
//    min(rev) in δFtail is min(@at) at which head/bigfile/file is currently mmapped (see below).
//
//    to support initial openings with @at being slightly in the past, we also
//    make sure that min(rev) is enough to cover last 10 minutes of history
//    from head/at.
//
// 7) when we receive a FUSE read(#blk) request to a head/bigfile/file we process it as follows:
//
//   7.1) load blkdata for head/bigfile/file[blk] @zhead.at .
//
//	while loading this also gives upper bound estimate of when the block
//	was last changed:
//
//	  rev(blk) ≤ max(_.serial for _ in (ZBlk(#blk), all BTree/Bucket that lead to ZBlk))
//
//	it is not exact because BTree/Bucket can change (e.g. rebalance)
//	but still point to the same k->ZBlk.
//
//	we also use file.δFtail to find either exact blk revision:
//
//	  rev(blk) = max(file.δFtail.by(#blk) -> []rev↑)
//
//	or another upper bound if #blk ∉ δFtail:
//
//	  rev(blk) ≤ min(rev ∈ δFtail)		; #blk ∉ δFtail
//
//
//	below rev'(blk) is min(of the estimates found):
//
//	  rev(blk) ≤ rev'(blk)		rev'(blk) = min(^^^)
//
//
//   7.2) for all client@at mmappings of head/bigfile/file:
//
//	- rev'(blk) ≤ at: -> do nothing
//	- rev'(blk) > at:
//	  - if blk ∈ mmapping.pinned -> do nothing
//	  - rev = max(δFtail.by(#blk) : _ ≤ at)	|| min(rev ∈ δFtail : rev ≤ at)	|| at
//	  - client.remmap(file, #blk, @rev/bigfile/file)
//	  - mmapping.pinned += blk
//
//	remmapping is done via "invalidation protocol" exchange with client.
//	( one could imagine adjusting mappings synchronously via running
//	  wcfs-trusted code via ptrace that wcfs injects into clients, but ptrace
//	  won't work when client thread is blocked under pagefault or syscall(^) )
//
//	in order to support remmapping for each head/bigfile/file
//
//	  [] of mmapping{client@at↑, pinned}
//
//	is maintained.
//
//   7.3) blkdata is returned to kernel.
//
//   Thus a client that wants latest data on pagefault will get latest data,
//   and a client that wants @rev data will get @rev data, even if it was this
//   "old" client that triggered the pagefault(~).
//
// (*) see notes.txt -> "Notes on OS pagecache control"
// (+) see notes.txt -> "Invalidations to wcfs clients are delayed until block access"
// (~) see notes.txt -> "Changing mmapping while under pagefault is possible"
// (^) see notes.txt -> "Client cannot be ptraced while under pagefault"
// (%) no need to keep track of ZData - ZBlk1 is always marked as changed on blk data change.
//
// XXX 8) serving read from @<rev>/data + zconn(s) for historical state
//
// XXX For every ZODB connection a dedicated read-only transaction is maintained.

import (
	"context"
	"flag"
	stdlog "log"
	"os"
	"strings"
	"sync"
	"syscall"

	log "github.com/golang/glog"
	"golang.org/x/sync/errgroup"

	"lab.nexedi.com/kirr/go123/xcontext"
	"lab.nexedi.com/kirr/go123/xerr"

	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
	_ "lab.nexedi.com/kirr/neo/go/zodb/wks"

	"github.com/hanwen/go-fuse/fuse"
	"github.com/hanwen/go-fuse/fuse/nodefs"
	"github.com/pkg/errors"
)

// Root represents root of wcfs filesystem.
type Root struct {
	nodefs.Node

	// ZODB storage we work with
	zstor zodb.IStorage

	// ZODB DB handle for zstor.
	// keeps cache of connections for both head/ and @<rev>/ accesses.
	//
	// only one connection is used for head/ and only one for each @<rev>.
	zdb *zodb.DB

	// ZODB connection for head/
	zheadMu sync.RWMutex // protects access to zhead & live _objects_ associated with it
	zhead   *ZConn       // zwatcher resyncs zhead; others only read zhead objects.

	// ZODB connections for @<rev>/
	zrevMu  sync.Mutex
	zrevTab map[zodb.Tid]*ZConn
}

// /bigfile/	- served by BigFileRoot.
type BigFileRoot struct {
	nodefs.Node

	// {} oid -> <bigfileX>/
	mu  sync.Mutex
	tab map[zodb.Oid]*BigFileDir
}

// /bigfile/<bigfileX>/	- served by BigFileDir.
type BigFileDir struct {
	nodefs.Node
	oid zodb.Oid // oid of ZBigFile

	// head/ is implicitly linked to by fs

	// {} rev -> @<rev>/ bigfile snapshot
	mu     sync.Mutex
	revTab map[zodb.Tid]*BigFileRev
}

// /bigfile/<bigfileX>/(head|<rev>)/	- served by BigFileRev.
type BigFileRev struct {
	nodefs.Node
	// data, at, invalidations, etc - all implicitly linked to by fs
}

// /bigfile/<bigfileX>/(head|<rev>)/*	- internally served by BigFile.
type BigFile struct {
	// this BigFile views ZODB via zconn
	zconn	*ZConn

	// ZBigFile top-level object. Kept activated during lifetime of current transaction.
	zbf	*ZBigFile

	// zbf.Size(). It is constant during liftime of current transaction.
	zbfSize int64

	// change history of this file.
	δFtail *ΔTailI64 // [](rev, []#blk)

	// TODO -> δFtail
	// lastChange	zodb.Tid // last change to whole bigfile as of .zconn.At view
}

// /bigfile/<bigfileX>/(head|<rev>)/data	- served by BigFileData.
type BigFileData struct {
	nodefs.Node

	bigfile *BigFile

	// inflight loadings of ZBigFile from ZODB.
	// successfull load results are kept here until blkdata is put into OS pagecache.
	//
	// XXX -> BigFile ?
	loadMu  sync.Mutex
	loading map[int64]*blkLoadState // #blk -> {... blkdata}

	// XXX mappings where client(s) requested isolation guarantee
	//mappings ...
}

// blkLoadState represents a ZBlk load state/result.
//
// when !ready the loading is in progress.
// when ready the loading has been completed.
type blkLoadState struct {
	ready chan struct{}

	blkdata []byte
	err     error
}

// ----------------------------------------

// zodbCacheControl implements zodb.LiveCacheControl to tune ZODB to never evict
// LOBTree/LOBucket from live cache. We want to keep LOBTree/LOBucket always alive
// becuse it is essentially the index where to find ZBigFile data.
//
// For the data itself - we put it to kernel pagecache and always deactivate
// from ZODB right after that.
//
// See "3) for */head/data the following invariant is maintained..."
type zodbCacheControl struct {}

func (cc *zodbCacheControl) WantEvict(obj zodb.IPersistent) bool {
	switch obj.(type) {
	default:
		return true

	case *btree.LOBTree:
	case *btree.LOBucket:

	// ZBlk* are kept referenced by a LOBucket, so they don't go away from Connection.cache.objtab

	// we also keep ZBigFile alive because we want to make sure .blksize
	// and (p. ref) .blktab do not change.
	// XXX do we really need to keep ZBigFile alive for that?
	//case *ZBigFile:
	}

	return false
}

// zwatcher watches for ZODB changes.
// see "4) when we receive an invalidation message from ZODB ..."
func (r *Root) zwatcher(ctx context.Context) (err error) {
	defer xerr.Contextf(&err, "zwatch")	// XXX more in context?
	// XXX unmount on error? -> always EIO?

	zwatch := r.zstor.Watch() // XXX -> create before zwatcher is started

	for {
		zevent, err := zwatch.Read(ctx)
		if err != nil {
			return err
		}

		r.zhandle1(zevent)
	}
}

// zhandle1 handles 1 event from ZODB notification.
// (called with .zheadMu wlocked)
func (r *Root) zhandle1(zevent zodb.WatchEvent) {
	// XXX locking correct? XXX too coarse? -> lock only around "resync .zhead ..." ?
	r.zheadMu.Lock()
	defer r.zheadMu.Unlock()

	//toinvalidate := map[*ZBigFile]SetI64{} // {} zfile -> set(#blk)
	toinvalidate := map[*BigFileData]SetI64{} // {} zfile -> set(#blk)

	// zevent = (tid^, []oid)
	for _, oid := range zevent.Changev {
		obj := r.zhead.Cache().Get(oid)
		if obj == nil {
			continue // nothing to do - see invariant
		}

		switch obj := obj.(type) {
		default:
			continue // object not related to any bigfile

		case *btree.LOBTree:
			// XXX -> δBTree

		case *btree.LOBucket:
			// XXX -> δBTree

		case zBlk:	// ZBlk*
			// XXX locking ?
			for zfile, objWhere := range obj.inzfile {
				blkmap, ok := toinvalidate[zfile]
				if !ok {
					blkmap = SetI64{}
					toinvalidate[zfile] = blkmap
				}
				blkmap.Update(objWhere)
			}

		case *ZBigFile:
			// XXX check that .blksize and .blktab (it is only
			// persistent reference) do not change.

			// XXX shutdown fs with ^^^ message.
		}
	}

	//wg = ...
	ctx := context.TODO()
	for file, blkmap := range toinvalidate {
		for blk := range blkmap {
			go file.invalidateBlk(ctx, blk)	// XXX -> wg.Go
		}
	}

	// XXX resync .zhead to zevent.tid
}

// invalidateBlk invalidates 1 file block.	XXX
// XXX see "4.4) for all file/blk to in invalidate we do"
func (f *BigFileData) invalidateBlk(ctx context.Context, blk int64) error {
	fsconn := f.root().fsconn
	off := blk*blksize

	// try retrieve cache of current head/data[blk]
	//
	// if less than blksize was cached - probably the kernel had to evict
	// some data from its cache already. In such case we don't try to
	// preserve the rest and drop what was read, to avoid keeping the
	// system overloaded.
	//
	// XXX st != OK -> warn?
	blkdata, st := fsconn.FileRetrieveCache(f.Inode(), off, blksize)
	if len(blkdata) == blksize {
		// XXX -> go
		// store retrieved data back to OS cache for file @<rev>/data[blk]
		frev, _ := f.bigfile.δFtail.LastRevOf(blk, at)
		st = fsconn.FileNotifyStoreCache(frev.Inode(), off, blkdata)
		if st != fuse.OK {
			// XXX log	- dup wrt readBlk -> common func.
		}
	}

	// invalidate file/head/data[blk] in OS file cache.
	st = fsconn.FileNotify(f.Inode(), off, blksize)
	// XXX st != ok (fatal here)

	panic("TODO")
}

// ----------------------------------------

// /bigfile -> Mkdir receives client request to create /bigfile/<bigfileX>.
//
// It creates <bigfileX>/head/* along the way.
func (bfroot *BigFileRoot) Mkdir(name string, mode uint32, fctx *fuse.Context) (*nodefs.Inode, fuse.Status) {
	inode, err := bfroot.mkdir(name, fctx) // XXX ok to ignore mode?
	return inode, err2LogStatus(err)

}

func (bfroot *BigFileRoot) mkdir(name string, fctx *fuse.Context) (_ *nodefs.Inode, err error) {
	defer xerr.Contextf(&err, "/bigfile: mkdir %q", name)

	oid, err := zodb.ParseOid(name)
	if err != nil {
		return nil, eINVALf("not oid")
	}

	// check to see if dir(oid) is already there
	bfroot.mu.Lock()
	_, already := bfroot.tab[oid]
	bfroot.mu.Unlock()

	if already {
		return nil, syscall.EEXIST
	}

	// not there - without bfroot lock proceed to open BigFile from ZODB
	bf, err := bigopen(asctx(fctx), groot.zhead, oid)
	if err != nil {
		return nil, err
	}
	defer func() {
		if err != nil {
			bf.Close()
		}
	}()

	// relock bfroot and either mkdir or EEXIST if the directory was maybe
	// simultanously created while we were not holding bfroot.mu
	bfroot.mu.Lock()
	_, already = bfroot.tab[oid]
	if already {
		bfroot.mu.Unlock()
		return nil, syscall.EEXIST
	}

	bfdir := &BigFileDir{
		Node:   nodefs.NewDefaultNode(),
		oid:    oid,
		revTab: make(map[zodb.Tid]*BigFileRev),
	}

	bfhead := &BigFileRev{
		Node: nodefs.NewDefaultNode(),
	}

	bfdata := &BigFileData{
		Node:    nodefs.NewDefaultNode(),
		bigfile: bf,
		loading: make(map[int64]*blkLoadState),
	}

	bfroot.tab[oid] = bfdir
	bfroot.mu.Unlock()

	// mkdir takes filesystem treeLock - do it outside bfroot.mu
	mkdir(bfroot, name, bfdir)
	mkdir(bfdir, "head", bfhead)
	mkfile(bfhead, "data", bfdata)
	mkfile(bfhead, "at", NewSmallFile(bf.readAt))	// TODO mtime(at) = tidtime(at)
	// XXX mkfile(bh, "invalidations", bh.inv)

	return bfdir.Inode(), nil
}

// XXX do we need to support rmdir? (probably no)

// /bigfile/<bigfileX> -> Mkdir receives client request to create @<tid>/.
func (bfdir *BigFileDir) Mkdir(name string, mode uint32, fctx *fuse.Context) (*nodefs.Inode, fuse.Status) {
	inode, err := bfdir.mkdir(name, fctx) // XXX ok to ignore mode?
	return inode, err2LogStatus(err)
}

func (bfdir *BigFileDir) mkdir(name string, fctx *fuse.Context) (_ *nodefs.Inode, err error) {
	defer xerr.Contextf(&err, "/bigfile/%s: mkdir %q", bfdir.oid, name)

	var tid zodb.Tid
	ok := false

	if strings.HasPrefix(name, "@") {
		tid, err = zodb.ParseTid(name[1:])
		ok = (err == nil)
	}
	if !ok {
		return nil, eINVALf("not @tid")
	}

	// check to see if dir(tid) is already there
	bfdir.mu.Lock()
	_, already := bfdir.revTab[tid]
	bfdir.mu.Unlock()

	if already {
		return nil, syscall.EEXIST
	}

	// not there - without bfdir lock proceed to open BigFile @tid view of ZODB
	ctx := asctx(fctx)
	zconnRev, err := groot.zopenAt(ctx, tid)
	if err != nil {
		return nil, err
	}
	defer zconnRev.Release()

	bf, err := bigopen(ctx, zconnRev, bfdir.oid)
	if err != nil {
		return nil, err
	}
	defer func() {
		if err != nil {
			bf.Close()
		}
	}()

	// relock bfdir and either mkdir or EEXIST if the directory was maybe
	// simultanously created while we were not holding bfroot.mu
	bfdir.mu.Lock()
	_, already = bfdir.revTab[tid]
	if already {
		bfdir.mu.Unlock()
		return nil, syscall.EEXIST
	}

	bfrev := &BigFileRev{
		Node: nodefs.NewDefaultNode(),
	}

	revdata := &BigFileData{
		Node:    nodefs.NewDefaultNode(),
		bigfile: bf,
		loading: make(map[int64]*blkLoadState),
	}

	bfdir.revTab[tid] = bfrev
	bfdir.mu.Unlock()

	// mkdir takes filesystem treeLock - do it outside bfroot.mu
	mkdir(bfdir, name, bfrev)
	mkfile(bfrev, "data", revdata)

	return bfrev.Inode(), nil
}


// bigopen opens BigFile corresponding to oid on zconn.
//
// A ZBigFile corresponding to oid is activated and statted.
//
// The whole result is returned as BigFile.
func bigopen(ctx context.Context, zconn *ZConn, oid zodb.Oid) (_ *BigFile, err error) {
	defer xerr.Contextf(&err, "bigopen %s @%s", oid, zconn.At())

	// XXX better ctx = transaction.PutIntoContext(ctx, txn)
	ctx, cancel := xcontext.Merge(ctx, zconn.txnCtx)
	defer cancel()

	xzbf, err := zconn.Get(ctx, oid)
	if err != nil {
		switch errors.Cause(err).(type) {
		case *zodb.NoObjectError:
			return nil, eINVAL(err)
		case *zodb.NoDataError:
			return nil, eINVAL(err) // XXX what to do if it was existing and got deleted?
		default:
			return nil, err
		}
	}

	zbf, ok := xzbf.(*ZBigFile)
	if !ok {
		return nil, eINVALf("%s is not a ZBigFile", typeOf(xzbf))
	}

	// activate ZBigFile and keep it this way
	err = zbf.PActivate(ctx)
	if err != nil {
		return nil, err
	}
	defer func() {
		if err != nil {
			zbf.PDeactivate()
		}
	}()

	zbfSize, err := zbf.Size(ctx)
	if err != nil {
		return nil, err
	}

	zconn.Incref()
	return &BigFile{
		zconn:   zconn,
		zbf:     zbf,
		zbfSize: zbfSize,

		// XXX this is needed only for head/
		δFtail:  NewΔTailI64(),	// XXX indicate we have coverage starting from zconn.at?
	}, nil
}

// Close release all resources of BigFile.
func (bf *BigFile) Close() error {
	bf.zbf.PDeactivate()
	bf.zbf = nil

	bf.zconn.Release()
	bf.zconn = nil

	return nil
}

// /bigfile/<bigfileX>/head/data -> Getattr serves stat.
func (bfdata *BigFileData) GetAttr(out *fuse.Attr, _ nodefs.File, fctx *fuse.Context) fuse.Status {
	// XXX locking

	bf := bfdata.bigfile

	out.Mode = fuse.S_IFREG | 0444
	out.Size = uint64(bf.zbfSize)
	// .Blocks
	// .Blksize

	// FIXME lastChange should cover all bigfile data, not only ZBigFile itself
	//mtime := &bfdata.lastChange.Time().Time
	lastChange := bf.zbf.PSerial()

	mtime := lastChange.Time().Time
	out.SetTimes(/*atime=*/nil, /*mtime=*/&mtime, /*ctime=*/&mtime)

	return fuse.OK
}


// /bigfile/<bigfileX>/head/data -> Read serves reading bigfile data.
func (bfdata *BigFileData) Read(_ nodefs.File, dest []byte, off int64, fctx *fuse.Context) (fuse.ReadResult, fuse.Status) {
	// XXX locking

	bf := bfdata.bigfile
	zbf := bf.zbf

	// cap read request to file size
	end := off + int64(len(dest))		// XXX overflow?
	if end > bf.zbfSize {
		end = bf.zbfSize
	}
	if end <= off {
		// XXX off >= size -> EINVAL? (but when size=0 kernel issues e.g. [0 +4K) read)
		return fuse.ReadResultData(nil), fuse.OK
	}

	// widen read request to be aligned with blksize granularity
	// (we can load only whole ZBlk* blocks)
	aoff := off - (off % zbf.blksize)
	aend := end
	if re := end % zbf.blksize; re != 0 {
		aend += zbf.blksize - re
	}
	dest = make([]byte, aend - aoff) // ~> [aoff:aend) in file

	// XXX better ctx = transaction.PutIntoContext(ctx, txn)
	ctx, cancel := xcontext.Merge(asctx(fctx), bf.zconn.txnCtx)
	defer cancel()

	// read/load all block(s) in parallel
	wg, ctx := errgroup.WithContext(ctx)
	for blkoff := aoff; blkoff < aend; blkoff += zbf.blksize {
		blkoff := blkoff
		blk := blkoff / zbf.blksize
		wg.Go(func() error {
			δ := blkoff-aoff // blk position in dest
			//log.Infof("readBlk #%d dest[%d:+%d]", blk, δ, zbf.blksize)
			return bfdata.readBlk(ctx, blk, dest[δ:δ+zbf.blksize])
		})
	}

	err := wg.Wait()
	if err != nil {
		log.Errorf("%s", err)	// XXX + /bigfile/XXX: read [a,b): -> ...
		return nil, fuse.EIO
	}

	return fuse.ReadResultData(dest[off-aoff:end-aoff]), fuse.OK
}

// readBlk serves Read to read 1 ZBlk #blk into destination buffer.
//
// see "6) when we receive a FUSE read(#blk) request ..." in overview.
//
// len(dest) == blksize.
func (bfdata *BigFileData) readBlk(ctx context.Context, blk int64, dest []byte) error {
	// XXX errctx?
	// XXX locking

	// check if someone else is already loading this block
	bfdata.loadMu.Lock()
	loading, already := bfdata.loading[blk]
	if !already {
		loading = &blkLoadState{
			ready:   make(chan struct{}),
		}
		bfdata.loading[blk] = loading
	}
	bfdata.loadMu.Unlock()

	// if it is already loading - just wait for it
	if already {
		select {
		case <-ctx.Done():
			return ctx.Err()

		case <-loading.ready:
			if loading.err == nil {
				copy(dest, loading.blkdata)
			}
			return loading.err
		}
	}

	// noone was loading - we became reponsible to load this block

	zbf := bfdata.bigfile.zbf
	blkdata, err := zbf.LoadBlk(ctx, blk)	// XXX -> +blkrevmax1
	loading.blkdata = blkdata
	loading.err = err
	close(loading.ready)

	// XXX before loading.ready?
	blkrevmax2, _ := bfdata.bigfile.δFtail.LastRevOf(blk, zbf.PJar().At())
	//revmax := min(blkrevmax1, blkrevmax2)
	revmax := blkrevmax2
	_ = revmax

/*
	// XXX remmapping
	// XXX -> own func?
	// XXX locking
	for _, mapping := range bfdata.mappings {
		if revmax <= mapping.at || !mapping.blkrange.in(blk) {
			continue // do nothing
		}

		if mapping.pinned.Contains(blk) {
			continue // do nothing
		}

		rev = max(δFtail.by(blk) : _ <= mapping.at)

		// XXX vvv -> go
		client.remmap(mapping.addr[blk], file/@<rev>/data)
		mapping.pinned.Add(blk)


	}
*/


	// data loaded with error - cleanup .loading
	if loading.err != nil {
		bfdata.loadMu.Lock()
		delete(bfdata.loading, blk)
		bfdata.loadMu.Unlock()
		return err
	}

	// data loaded ok
	copy(dest, blkdata)

	// store to kernel pagecache whole block that we've just loaded from database.
	// This way, even if the user currently requested to read only small portion from it,
	// it will prevent next e.g. consecutive user read request to again hit
	// the DB, and instead will be served by kernel from its pagecache.
	//
	// We cannot do this directly from reading goroutine - while reading
	// kernel FUSE is holding corresponging page in pagecache locked, and if
	// we would try to update that same page in pagecache it would result
	// in deadlock inside kernel.
	//
	// .loading cleanup is done once we are finished with putting the data into OS pagecache.
	// If we do it earlier - a simultaneous read covered by the same block could result
	// into missing both kernel pagecache (if not yet updated) and empty .loading[blk],
	// and thus would trigger DB access again.
	go func() {
		// XXX locking - invalidation must make sure this workers are finished.

		// XXX if direct-io: don't touch pagecache
		st := gfsconn.FileNotifyStoreCache(bfdata.Inode(), blk*zbf.blksize, blkdata)

		bfdata.loadMu.Lock()
		delete(bfdata.loading, blk)
		bfdata.loadMu.Unlock()

		if st == fuse.OK {
			return
		}

		// pagecache update failed, but it must not (we verified on startup that
		// pagecache control is supported by kernel). We can correctly live on
		// with the error, but data access will be likely very slow. Tell user
		// about the problem.
		log.Errorf("BUG: bigfile %s: blk %d: -> pagecache: %s  (ignoring, but reading from bigfile will be very slow)", zbf.POid(), blk, st)
	}()

	return nil
}


// /bigfile/<bigfileX>/head/at -> readAt serves read.
func (bf *BigFile) readAt() []byte {
	// XXX locking
	// XXX zbf.PJar() not good if we want to share objects between connections?
	return []byte(bf.zbf.PJar().At().String())
}





// FIXME groot/gfsconn is tmp workaround for lack of way to retrieve FileSystemConnector from nodefs.Inode
// TODO:
//	- Inode += .Mount() -> nodefs.Mount
//	- Mount:
//		.Root()		-> root Inode of the fs
//		.Connector()	-> FileSystemConnector through which fs is mounted
var groot   *Root
var gfsconn *nodefs.FileSystemConnector

func main() {
	stdlog.SetPrefix("wcfs: ")
	//log.CopyStandardLogTo("WARNING") // XXX -> "DEBUG" if -d ?
	defer log.Flush()

	debug := flag.Bool("d", false, "debug")
	autoexit := flag.Bool("autoexit", false, "automatically stop service when there is no client activity")
	// XXX option to prevent starting if wcfs was already started ?

	flag.Parse()
	if len(flag.Args()) != 2 {
		log.Fatalf("Usage: %s [OPTIONS] zurl mntpt", os.Args[0])
	}
	zurl := flag.Args()[0]
	mntpt := flag.Args()[1]

	// debug -> precise t, no dates	(XXX -> always precise t?)
	if *debug {
		stdlog.SetFlags(stdlog.Lmicroseconds)
	}

	// open zodb storage/db/connection
	ctx := context.Background()	// XXX + timeout?
	zstor, err := zodb.OpenStorage(ctx, zurl, &zodb.OpenOptions{ReadOnly: true})
	if err != nil {
		log.Fatal(err)
	}
	defer zstor.Close()

	zdb := zodb.NewDB(zstor)
	zhead, err := zopen(ctx, zdb, &zodb.ConnOptions{})
	if err != nil {
		log.Fatal(err)
	}
	zhead.Cache().SetControl(&zodbCacheControl{})	// XXX +locking?

	// mount root
	root := &Root{
		Node:    nodefs.NewDefaultNode(),
		zstor:   zstor,
		zdb:     zdb,
		zhead:   zhead,
		zrevTab: make(map[zodb.Tid]*ZConn),
	}

	opts := &fuse.MountOptions{
		FsName: zurl,
		Name:   "wcfs",

		DisableXAttrs: true, // we don't use
		Debug:         *debug,
	}

	fssrv, fsconn, err := mount(mntpt, root, opts)
	if err != nil {
		log.Fatal(err)
	}
	groot   = root		// FIXME temp workaround (see ^^^)
	gfsconn = fsconn	// FIXME ----//----

	// we require proper pagecache control (added to Linux 2.6.36 in 2010)
	supports := fssrv.KernelSettings().SupportsNotify
	if !(supports(fuse.NOTIFY_STORE_CACHE) && supports(fuse.NOTIFY_RETRIEVE_CACHE)) {
		log.Fatalf("kernel FUSE does not support pagecache control")
	}

	// add entries to /
	mkfile(root, ".wcfs", NewStaticFile([]byte(zurl)))
	mkdir(root, "bigfile", &BigFileRoot{
		Node:  nodefs.NewDefaultNode(),
		tab:   make(map[zodb.Oid]*BigFileDir),
	})

	// TODO handle autoexit
	_ = autoexit

	// serve client requests
	fssrv.Serve()	// XXX Serve returns no error
}
